#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: sql.tcl,v 1.13 2000/10/20 09:44:20 artur Exp $
#
# Information:
# Modul (benutzt von Schema Editor und Form Server)
# Diese Datei beinhaltet die Abstrakte XOTcl Klasse
# Sqlinterface
# und eine Implementierung als mysql API und PostgresSql Tcl Interface

Class Sqlinterface -parameter {dbank {connected 0}}
Sqlinterface abstract instproc getconnectargs {}
Sqlinterface abstract instproc connect par_list
Sqlinterface abstract instproc getdblist {}
Sqlinterface abstract instproc setdb dbname
Sqlinterface abstract instproc disconnect {}
Sqlinterface instproc destroy {} {
    # Vielleicht bracht das Klassenkind noch die Verbindung
    #next
    # Next verursacht ein Fehler +++ ::Cant find self
    [self] instvar connected
    if {$connected} {
	[self] disconnect
    }
    next
}
# Sqlinterface instproc connectfilter args {
#puts filtered
#     set needconnect {
# 	getdblist
# 	setdb
#     }
#     if {[lsearch $needconnect [[self] info calledproc]]!=-1} {
# 	[self] instvar connected
# 	if {$connected==0} { error {mysql not connected}}
#     }
#     [self] next
# }
#Sqlinterface filter connectfilter
Class MySql -superclass Sqlinterface -parameter {
    user
    password
    hostname
    {interface mysql}
}
MySql set interface mysql
# Instanz Variablen
# sqlhandle - Handle fr 
MySql instproc init args {
    package require mysqltcl
    next
}
MySql instproc getconnectargs {} {
    global env
    if {[info exist env(USERNAME)]} {
	set user $env(USERNAME)
    } else {
	set user ""
    }
    return [list \
		{hostname text localhost} \
		{port integer 3306} \
		[list user text $user] \
		{password password {}} \
		{dbank text {}}
	    ]
	   
}

MySql instproc connect par_list {
    [self] instvar sqlhandle_arr connected dbank user password hostname
    array set par_array [eval concat $par_list]
    #puts "checkpoint"
    set par ""
    if {[info exists par_array(user)] && $par_array(user)!=""} {
	lappend par -user $par_array(user) 
	set user $par_array(user) 
    }
    if {[info exists par_array(password)] && $par_array(password)!=""} {
	lappend par -password $par_array(password) 
	set password $par_array(password) 
    }
    if {[info exists par_array(hostname)] && $par_array(hostname)!=""} {
	lappend par $par_array(hostname) 
	set hostname $par_array(hostname) 
    }
    # Kein Tcl-Fehler Abfang
    #puts "$par $par_list"
    if {[catch {set sqlhandle [eval mysqlconnect $par]} err]!=0} {
	global errorInfo
	puts "$err $errorInfo"
	return 1
    }
    set connected 1
    set sqlhandle_arr($sqlhandle) free
    if {[info exists par_array(dbank)] && $par_array(dbank)!=""} {
	if {[catch {mysqluse $sqlhandle $par_array(dbank)} err]!=0} {
	    global errorInfo
	    puts "$err $errorInfo"
	    [self] disconnect
	    return 1
	}
	set dbank $par_array(dbank)
    }    
    return 0
}

MySql instproc disconnect {} {
    [self] instvar sqlhandle_arr connected handlelist
    foreach handle [array names sqlhandle_arr] {
	mysqlclose $handle
    }
    set connected 0
}
MySql instproc getdblist {} {
    [self] instvar sqlhandle_arr
    set  sqlhandle [lindex [array names sqlhandle_arr] 0]
    return [mysqlinfo $sqlhandle databases]
}
MySql instproc setdb dbname {
    [self] instvar sqlhandle_arr dbank
    set sqlhandle [lindex [array names sqlhandle_arr] 0]
    if {[catch {mysqluse $sqlhandle $dbname} err]!=0} {
	global errorInfo
	puts "$err $errorInfo"
	set dbank {}
	return
    }
    set dbank $dbname
}
MySql instproc freehandle handle {
    [self] instvar sqlhandle_arr
    set sqlhandle_arr($handle) free
}
# Strategie es werden immer neue Hanlde geffnet aber bei nicht bedarf nicht
# freigegeben (Taktik. jedes ffnen dauert zu lange)
MySql instproc gethandle {} {
    [self] instvar sqlhandle_arr
    foreach handle [array names sqlhandle_arr] {
	if {$sqlhandle_arr($handle)=="free"} {
	    return $handle
	}
    }
    return [[self] clonehandle]
}
# Nur fr interne Nutzung
# Gibt ein Hadle zurck auf die gleiche Datenbank und Tabele wie bei dbconnect
# Benutzt um mehrfach Anfragen zu erlauben
MySql instproc clonehandle {} {
    [self] instvar sqlhandle_arr dbank user password hostname
    set par ""
    if {[info exists user]} {
	lappend par -user $user
    }
    if {[info exists password]} {
	lappend par -password $password
    }
    if {[info exists hostname]} {
	lappend par $hostname
    }
    # Kein Tcl-Fehler Abfang
    #puts "clonehandle $par"
    set chandle [eval mysqlconnect $par]
    # Vorausetzung es wurde eine konkrete dbank bereits gesetzt
    mysqluse $chandle $dbank
    set sqlhandle_arr($chandle) free
    return $chandle
}
# Sendet eine SQL Anweisung an Server und gibt ein SqlQuerry Objekt
# Zurck, das fr die Tranfersiereung der Antwort (Tablle=Menge) 
# benutzt werden kann
MySql instproc query sqlstatment {
    [self] instvar sqlhandle_arr
    set qhandle [[self] gethandle]
    set sqlhandle_arr($qhandle) inuse
    set rows [mysqlsel $qhandle $sqlstatment]
    set name [Object autoname mySqlQuerry]
    SqlQuerry [self]::$name -qhandle $qhandle -frows $rows
    return [self]::$name
}
# Sendet eine SQL Anweisung an Server (keine SELECT)
MySql instproc execute sqlstatement {
    [self] instvar sqlhandle_arr
    set qhandle [[self] gethandle]
    return [mysqlexec $qhandle $sqlstatement]
}
MySql instproc getinsertid {} {
    [self] instvar sqlhandle_arr
    set qhandle [[self] gethandle]
    return [mysqlinsertid $qhandle]
}
Class SqlQuerry -parameter {
    {qhandle 0}
    {frows 0}
}

SqlQuerry instproc fetch {} {
    [self] instvar qhandle
    return [mysqlnext $qhandle]
}
SqlQuerry instproc rows {} {
    [self] instvar frows
    return $frows
}
SqlQuerry instproc seek {point} {
    [self] instvar qhandle
    mysqlseek $qhandle $point
}
SqlQuerry instproc destroy {} {
    [self] instvar qhandle 
    [[self] info parent] freehandle $qhandle
    next
}
#
# Interface fr Postgres SQL Datenbank
# 
Class PgSql -superclass Sqlinterface -parameter {
    user
    password
    hostname
    {interface postgresql}
}
PgSql set interface postgresql
# Instanz Variablen
# sqlhandle - Handle fr 
PgSql instproc init args {
    package require Pgtcl
    next
}
PgSql instproc getconnectargs {} {
    global env
    [self] instvar dbank
    foreach arg [pg_conndefaults] {
	if {[lindex $arg 0]=="password"} {
	    lappend connlist [list [lindex $arg 0] password [lindex $arg 4]]
	} else {
	    lappend connlist [list [lindex $arg 0] text [lindex $arg 4]]
	    if {[lindex $arg 0]=="dbname"} {
		set dbank [lindex $arg 4]
	    }
	}
    }
    return $connlist
}

PgSql instproc connect par_list {
    [self] instvar sqlhandle connected dbank user password hostname
    set conninfo {}
    foreach par [keylkeys par_list] {
	if {[keylget par_list $par]!=""} {
	    if {$par=="interface"} {
		continue
	    }
	    if {$par=="dbname"} {
		set dbank [keylget par_list dbname]
	    }
	    [self]  instvar $par
	    set $par [keylget par_list $par]
	    lappend conninfo "$par='[keylget par_list $par]'"
	}
    }
    #puts "PGSQL $conninfo"
    if {[catch {set sqlhandle [pg_connect -conninfo $conninfo]} err]!=0} {
	global errorInfo
	puts "$err $errorInfo"
	return 1
    }
    set connected 1
}

PgSql instproc disconnect {} {
    [self] instvar sqlhandle connected
    if {$connected} {
	pg_disconnect $sqlhandle
    }
    set connected 0
}
PgSql instproc getdblist {} {
    error "postgresql interface don't support this funktionality. You must know names of your databases"
}
PgSql instproc setdb dbname {
    error "postgresql interface don't support this funktionality. You must know names of your databases"
}
PgSql instproc query sqlstatment {
    [self] instvar sqlhandle
    set qhandle [pg_exec $sqlhandle $sqlstatment]
    set name [Object autoname mySqlQuerry]
    #puts "Querry Status [pg_result $qhandle -status]"
    PSqlQuerry [self]::$name -qhandle $qhandle
    return [self]::$name
}
# Sendet eine SQL Anweisung an Server (keine SELECT)
PgSql instproc execute sqlstatement {
    [self] instvar sqlhandle
    set qhandle [pg_exec $sqlhandle $sqlstatement]
    set status [pg_result $qhandle -status]
    if {$status=="PGRES_FATAL_ERROR"} {
	error "postgresql ERROR [pg_result $qhandle -error]"
    } else {
	#puts "Querry Execute Status $status"
    }
    pg_result $qhandle -clear
    return $status
}
Class PSqlQuerry -parameter {
    {qhandle}
    {tindex 0}
    {numTuples}
}
PSqlQuerry instproc init {args} {
    next
    [self] instvar numTuples qhandle
    set numTuples [pg_result $qhandle -numTuples]
}
PSqlQuerry instproc fetch {} {
    [self] instvar qhandle tindex numTuples
    if {$tindex<$numTuples} {
	set result [pg_result $qhandle -getTuple $tindex]
	incr tindex
    } else {
	set result {}
    }
    #puts "fetch $tindex $result"
    return $result
}
PSqlQuerry instproc getinsertid {} {
    [self] instvar qhandle
    return [pg_result $qhandle -oid]
}
PSqlQuerry instproc rows {} {
    [self] instvar numTuples
    return $numTuples
}
PSqlQuerry instproc seek {point} {
    [self] instvar tindex
    set tindex $point
}
PSqlQuerry instproc destroy {} {
    [self] instvar qhandle 
    pg_result $qhandle -clear
    next
}
# braucht geladenes Array lan fr internationalle untersttzung und
# widgetlib.tcl und dialog.tcl
# hat ein dialog fr sql Verbindung zu Kreiren und ein SqlInterface
# Object zurckzugeben oder nicht wenn milungen
Object ConnectCreator
ConnectCreator proc getconnect {name {cpar {}} {withSchemaExt 0}} {
    global lan
    #puts "cpar $cpar"
    set cpark [keylkeys cpar]
    #  Finde alle Interface Klassen
    foreach iclass [Sqlinterface info subclass] {
	keylset if_classes [$iclass set interface] $iclass
    }
    set interfaces [keylkeys if_classes]
    # berschreibe wenn interface bereits bestimmt
    if {[lcontain $cpark interface]} {
	set interfaces [keylget cpar interface]
    }
    # Benutzer soll interface bestimmen
    if {[llength $interfaces]>1} {
	set interfacelist {{type OComboBox} {name interface}}
	keylset interfacelist label $lan(connect.interfacelist)
	keylset interfacelist valuelist $interfaces
	#puts $interfacelist
	WidgetFactory make_inputs_keylist [self]::inputframe [list $interfacelist]
	ODialog interfacedialog -title $lan(connect.chooseinterface) \
	    -message $lan(connect.chooseinterface)\
	    -inputframe [self]::inputframe -type okcancel
	if {[interfacedialog execute]=="cancel"} { return {}}
	set value [[self]::inputframe get_values]
	set interface [keylget value interface]
	interfacedialog destroy
	[self]::inputframe destroy
    } else {
	set interface $interfaces
    }
    # Interface Objekt bilden
    set iclass [keylget if_classes $interface]
    if {$withSchemaExt} {
	# Hier wird vorausgesetzt das die Kindklasse die 
	# Schema-Auslese Operationen beinhaltet
	set iclass [$iclass info subclass]
    }
    $iclass $name
    # Dialog fr Attributte des Verbindung zu konkreten DB bilden
    set par [$name getconnectargs]
    foreach arg $par {
	set tpar {}
	# setzt die Reiehnfolge typ name label default
	lappend tpar [lindex $arg 1]
	lappend tpar [lindex $arg 0]
	lappend tpar $lan(connect.[lindex $arg 0])
	# Wenn argument in par schon vorgegeban dann setzte auf default
	if {[lsearch $cpar [lindex $arg 0]*]!=-1} {
	    lappend tpar [keylget cpar [lindex $arg 0]]
	    #puts "inherit [lindex $arg 0] [keylget cpar [lindex $arg 0]]"
	} else {
	    lappend tpar [lindex $arg 2]
	}
	lappend inputs $tpar
    }
    WidgetFactory make_simply_inputs cinput $inputs
    ODialog connectdialog -title $lan(connect.mysql) \
	-message $lan(connect.mysql)\
	-inputframe cinput -type okcancel
    if {[connectdialog execute]=="cancel"} { 
	$name destroy
	connectdialog destroy
	cinput destroy
	return {}
    }
    set par [cinput get_changes]
    # Varibale par beinhaltet nur dir Vernderungen
    # alle vorgaben dazugeben (Variable cpar)
    # Welche Werte m
    foreach inhkey $cpark {
	if {![lcontain [keylkeys par] $inhkey]} {
	    keylset par $inhkey [keylget cpar $inhkey]
	}
    }
    $name connect $par
    connectdialog destroy
    cinput destroy
    if {[$name set connected]==0} {
	$name destroy
	return {}
    }
    # Wurde die Datenbank Ausgewlt?
    if {![info exist ${name}::dbank] || [set ${name}::dbank]==""} {
	set interfacelist {{type OComboBox} {name interface}}
	keylset interfacelist label $lan(connect.dbanklist)
	keylset interfacelist valuelist [$name getdblist]
	#puts $interfacelist
	WidgetFactory make_inputs_keylist [self]::inputframe [list $interfacelist]
	ODialog interfacedialog \
	    -message $lan(connect.choosedbank)\
	    -inputframe [self]::inputframe -type okcancel
	if {[interfacedialog execute]=="cancel"} { return {}}
	set value [[self]::inputframe get_values]
	$name setdb [keylget value interface]
	if {[$name set dbank]==""} {
	    $name disconnect
	    $name destroy
	    return {}
	}
	interfacedialog destroy
	[self]::inputframe destroy
    }
    return $name
}
